
# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '0.1.29'
version = '0.1.29'
full_version = '0.1.29.dev1550246703'
dev_version = '1550246703'
release = False
if not release:
    version = full_version
